"""Welcome to the documentation and API reference for DLKoopman - a Python package for Koopman theory using deep learning.

Source code: https://github.com/GaloisInc/dlkoopman

PyPI page: https://pypi.org/project/dlkoopman/
"""
